package polyglot.ext.op;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.frontend.*;
import polyglot.visit.*;
import java.util.*;

public class OPCastRewriter extends AscriptionVisitor
{
    public OPCastRewriter(Job job, TypeSystem ts, NodeFactory nf) {
        super(job, ts, nf);
    }

    public Expr ascribe(Expr e, Type toType) {
        Type fromType = e.type();

        if (toType == null) {
            return e;
        }

        Position p = e.position();

        // Insert a cast.  Translation of the cast will insert the
        // correct boxing/unboxing code.
        if (toType.isReference() && fromType.isPrimitive()) {
            return nf.Cast(p, nf.CanonicalTypeNode(p, ts.Object()), e);
        }

        return e;
    }

    public Node leaveCall(Node old, Node n, NodeVisitor v) throws SemanticException {
        n = super.leaveCall(old, n, v);

        // Insert boxing and unboxing code.
        if (n instanceof Cast) {
            Cast c = (Cast) n;
            Type rtype = c.expr().type();
            Type ltype = c.castType().type();
            TypeSystem ts = typeSystem();

            if (ltype.isPrimitive() && rtype.isReference()) {
                MethodInstance mi = ((OPTypeSystem) ts).getter(ltype.toPrimitive());

                Cast x = nf.Cast(c.position(),
                                 nf.CanonicalTypeNode(c.position(),
                                                      mi.container()),
                                 c.expr());
                x = (Cast) x.type(mi.container());

                Call y = nf.Call(c.position(), x, mi.name(),
                                 Collections.EMPTY_LIST);
                y = (Call) y.type(mi.returnType());

                return y.methodInstance(mi);
            }
            else if (ltype.isReference() && rtype.isPrimitive()) {
                ConstructorInstance ci = ((OPTypeSystem) ts).wrapper(rtype.toPrimitive());

                List args = new ArrayList(1);
                args.add(c.expr());

                New x = nf.New(c.position(),
                               nf.CanonicalTypeNode(c.position(),
                                                    ci.container()), args);
                x = (New) x.type(ci.container());
                return x.constructorInstance(ci);
            }
            else {
                return n;
            }
        }

        // Rewrite instanceof for primitives
        if (n instanceof Instanceof) {
            Instanceof o = (Instanceof) n;
            Type rtype = o.compareType().type();

            if (rtype.isPrimitive()) {
                ConstructorInstance ci = ((OPTypeSystem) ts).wrapper(rtype.toPrimitive());
                return o.compareType(nf.CanonicalTypeNode(o.compareType().position(),
                                                          ci.container()));
            }
        }

        // TODO: Rewrite == and != to unbox.
        if (n instanceof Binary) {
            Binary b = (Binary) n;

            // when l and r types are Object or <: Primitive

            if (b.operator() == Binary.EQ) {
                // (l == r || ((l instanceof Primitive) && l.equals(r)))
            }

            if (b.operator() == Binary.NE) {
                // ! (l == r || ((l instanceof Primitive) && l.equals(r)))
            }
        }

        return n;
    }
}
