package polyglot.ext.polyj.ast;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.util.*;
import polyglot.ext.polyj.visit.*;

import polyglot.ext.jl.ast.*;

import java.util.*;


/**
 * A <code>PolyJInstanceof</code> is an immutable representation of
 * the use of the <code>instanceof</code> operator in PolyJ programs.
 **/
public class PolyJInstanceof_c extends Instanceof_c implements PolyJInstanceof {

    /**
     * Creates a new <code>PolyJInstanceof_c</code>.
     */
    public PolyJInstanceof_c( Position pos, Expr expr, TypeNode compareType) {
        super(pos, expr, compareType);
    }
    
    /**
     * <code>instanceof</code> operators are translated to calls to methods 
     * for some parametrized types
     */
    public void translate(CodeWriter w, Translator tr) {
        PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();
        Context c = tr.context();
        Type castTo = compareType.type();

        if (castTo.isClass() && ts.needsWhereObject(castTo.toClass())) {
            ClassType classCast = castTo.toClass();
            w.write(PolyJTranslator.adapterInterfaceShortName(classCast));
            w.write(".$instanceof" + "(");

            printSubExpr(expr, w, tr);

            List typeArgs = ts.allTypeArgs(classCast);
            for (Iterator i = typeArgs.iterator(); i.hasNext(); ) {
                Type t = (Type) i.next();
                w.write(",\"" + t + "\"");
            }

            w.write(")");
        } else {
            printSubExpr(expr, w, tr);
            w.write(" instanceof ");
            compareType.del().translate(w, tr);
        }
    }
}
